#!/usr/bin/env python
 # ***************************************************************************
 #
 #                              dsh
 #
 # ***************************************************************************
 #
 # Copyright INRIA
 #  contributors :  Laurent Bulteau
 #                  Alice Julien-Laferriere
 #                  Delphine Parrot
 #
 # l.bulteau@gmail.com
 # alice.julien.laferriere@gmail.com
 #
 # This software is a computer program whose purpose is detect topological
 # feasible consortium of microorganisms for the production of compounds 
 # of interest
 #
 # This software is governed by the CeCILL license under French law and
 # abiding by the rules of distribution of free software. You can  use,
 # modify and/ or redistribute the software under the terms of the CeCILL
 # license as circulated by CEA, CNRS and INRIA at the following URL
 # "http://www.cecill.info".

 # As a counterpart to the access to the source code and  rights to copy,
 # modify and redistribute granted by the license, users are provided only
 # with a limited warranty  and the software's author,  the holder of the
 # economic rights,  and the successive licensors  have only  limited
 # liability.

 # In this respect, the user's attention is drawn to the risks associated
 # with loading,  using,  modifying and/or developing or reproducing the
 # software by the user in light of its specific status of free software,
 # that may mean  that it is complicated to manipulate,  and  that  also
 # therefore means  that it is reserved for developers  and  experienced
 # professionals having in-depth computer knowledge. Users are therefore
 # encouraged to load and test the software's suitability as regards their
 # requirements in conditions enabling the security of their systems and/or
 # data to be ensured and,  more generally, to use and operate it in the
 # same conditions as regards security.
 #
 # The fact that you are presently reading this means that you have had
 # knowledge of the CeCILL license and that you accept its terms.

import os 
import sys
import argparse
from subprocess import Popen, PIPE, STDOUT, call
from configparser import ConfigParser
import shlex
# ===================== UTILS =================================
def check_executable( name, newpath=""):
    name = newpath + name
    error  = call("type " + name, shell=True, stdout=PIPE, stderr=PIPE)
    if error !=  0:
        print "%s exec not found on path" % name
        sys.exit(1)

def launchCommandLine( commandLine, myEnv, outputFile, errorFile, verbose=False):
    if verbose:
        print commandLine, "running..."
  
    sys.stdout.flush() 
    stdoutput = open( outputFile, 'w')
    errorF = open( errorFile, 'w')
    sub = Popen(commandLine, stdout = stdoutput, stderr= PIPE, env = myEnv, shell=True)
    sub.communicate()
    stdoutput.close()
    errorF.close()
        

def main():

    # ========================================================================
    #                        Manage command line arguments
    # ========================================================================
    parser = argparse.ArgumentParser(description = " ") #  
    # ------------------------------------------------------------------------
    #                            Define allowed options
    # ------------------------------------------------------------------------  
    
     # Mandatory
    parser.add_argument('--SbmlF',        action = "store", help = "the files to proceed with. Please put them between quotes, separated by space") # nb find another way?
    parser.add_argument('--cofactors',    action = "store", help = "a file containing the metabolites to ignore, those would be cofactors and ubiquitous compounds (one by line)", required = False) # nb find another way?
    parser.add_argument('--idsNet',       action = "store", help = "ids of the species used. By default A,B,C .... Please put them between quotes, separated by space", required = False) 
    parser.add_argument('--insertionNet', action = "store", help = "the SBML file of the species from which we can insert genes. Please put them between quotes, separated by space", required = False) # nb find another way?
    parser.add_argument('--sources',      action = "store", help = "the sources file  (metabolites used as source), one file only") 
    parser.add_argument('--targets',      action = "store", help = "the targets file  (metabolites used as target), one file only") 
    parser.add_argument("--maxK",          action = "store",    default = 3,    help = "max number of tentacular hyperarcs in the solution (default 3)")
    parser.add_argument("--config", action="store", default="configMultipus.ini", help = "config file to use")
   
    parser.add_argument("--version", action="version", version="%(prog)s 0.1.0")     
   
    # ------------------------------------------------------------------------
    #               Parse and interpret command line arguments
    # ------------------------------------------------------------------------
    options = parser.parse_args()           
    my_env = os.environ.copy()


    # reading configFile 
    config = ConfigParser()
    config.read_file( open(options.config))


    prepPath =  config["path2software"]["preprocessingPath"]
    multiPusPath = config["path2software"]["multiPusPath"]
    multiPusOnly = bool(int(config["path2software"]["multiPusOnly"]))
    # checking if path matchs
    
    if not multiPusOnly : 
        if  prepPath :
            my_env["PATH"] = my_env["PATH"] + ":" + prepPath 
        indexDir = check_executable( "dsh_preprocessing", prepPath )
    if multiPusPath:
        my_env["PATH"] = my_env["PATH"] + ":" + multiPusPath
    
    indexDir = check_executable( "multipus_main", multiPusPath)
    #=============== Directory creations
    outDir =  config["output"]["output"]
    if not os.path.exists( outDir ):
        os.mkdir( outDir )  

    
    
    # ------------------------------------------------------------------------
    #                Preprocessing
    # ------------------------------------------------------------------------ 
    print (not multiPusOnly)
    print (multiPusOnly)
    print "yea"
    if not multiPusOnly:
        # creating preprocessing output directory
        tmpD = config["output"]["tmpD"]
        if not os.path.exists( tmpD  ):
            os.mkdir( tmpD  )  
        ## weight?
        print tmpD
        print "doing the preprocessing, see results in " +  tmpD
        wEndo = config["preprocessing"]["Wendo"]
        wInsert = config["preprocessing"]["Winsert"]
        wTransi = config["preprocessing"]["Wtransi"]

        commandLine = "dsh_preprocessing "  + "\"" + (options.SbmlF ) + "\""
        if options.cofactors:
            commandLine += " --cofactors " + options.cofactors
        if options.idsNet :
            commandLine += " --idsNet " + options.idsNet
        if options.insertionNet : 
            commandLine += " --insertionNet " + options.insertionNet
        commandLine += " --sources " + options.sources
        commandLine += " --targets " + options.targets
        commandLine += " --Wendo " + wEndo + " --Wtransi " + wTransi + " --Winsert " + wInsert + " --outputD " +  tmpD
        print "\n launching preprocessing... "
        launchCommandLine( commandLine, my_env, outputFile = tmpD + "/" + "preprocessingMultipus.OU", errorFile = tmpD + "/" +"preprocessingMultipus.ER", verbose = True) 
    
    else : 
        tmpD = config["path2software"]["multiFiles"]

    # ------------------------------------------------------------------------
    #               Multipus_core
    # ------------------------------------------------------------------------ 
    print tmpD
    multiPusParam = config["MultiPusParameters"]
    commandLine2 ="multipus_main --arcFile " + tmpD + "/finalFiles_edges.txt --hyperArcFile " +  tmpD + "/finalFiles_HyperEdges.txt "
    commandLine2 +=" --nbThreads "   + multiPusParam["nbThreads"] + " --reset "  + multiPusParam["reset"]  + " --reset_From  " + "\"" + multiPusParam["resetFrom"]+ "\""
    commandLine2 +=" --verboseLevel "+ multiPusParam["verbose"]   + " --dryrun " + multiPusParam["dryrun"] + " --USE_LB "      +  multiPusParam["UseLb"]+ " --KeepStats  " + multiPusParam["keepStats"]
    commandLine2 +=" --stepP " + multiPusParam["stepP"] + " --CHECK_ST " +  multiPusParam["checkST"]
    print "\n\n\n launching MultiPus... " 
    launchCommandLine( commandLine2, my_env, outputFile = outDir + "/Multipus.OU", errorFile = outDir + "/Multipus.ER", verbose = True) 
    print "\n End of Multipus, go see the results in :" + outDir



if __name__ == "__main__":
    main()
