
#include "Graph.h"




bool operator <(const HyperArc&  x, const HyperArc y)
{
    return  x.id < y.id;
}


//TODO : check that this path actually exists (otherwise it triggers memory faults).
std::vector< vertex_t > getShortestPath(nameMap name , map< vertex_t, vector<vertex_t> > allPredecessors, vertex_t startNode, vertex_t endNode, bool verbose )
{

    vector<vertex_t> predecessors  = allPredecessors.at(startNode);    //predecessors is the predecessor map obtained through dijkstra

    //VertexIndexMap index_map = get(vertex_index, graph); // ridiculous, we use it all the time...
    std::vector< vertex_t > path;
    vertex_t current= endNode;


    while(current!=startNode) {
        path.push_back(current);
        current=predecessors[current];
    }
    path.push_back(startNode);

    if(verbose)
    {
        //This prints the path reversed use reverse_iterator and rbegin/rend

      //  nameMap name2 = boost::get(&VertexProperties::name, graph);
        cout << "from " << name[startNode] << " to " << name[endNode] << endl;
        std::vector< vertex_t>::iterator it;
        for (it=path.begin(); it != path.end(); ++it) {

            std::cout << name[*it] << " ";
        }
        std::cout << std::endl;
    }
    return path;
}




/*!
 * \brief This function will fill distancesM and return for all vertices the predecessors map .
 * Beware distancesM must be initialized
 */
map< vertex_t, vector<vertex_t> > computeDijkstra( Graph g, numeric::ublas::matrix<double> & distancesM  )
                {
					cout << "Computing shortest path matrix" << endl;
    VertexIndexMap index_map = get(vertex_index, g);

    map< vertex_t, vector<vertex_t> > allPredecessors;
    //Predecessor Map
    vector<vertex_t> predecessors   = vector<vertex_t>(num_vertices(g));

    auto weightMap  = get(&EdgeProperties::weight, g); // get the weight
    //Distance MAp
    vector<double> distances =  vector<double>(num_vertices(g));

    pair<vertex_iter, vertex_iter> vp;
    for (vp = vertices(g); vp.first != vp.second; ++vp.first)
    {
        vertex_t beginV = *vp.first;
        dijkstra_shortest_paths( g, beginV,
            predecessor_map(boost::make_iterator_property_map(predecessors.begin(), get(boost::vertex_index, g)))
            .distance_map(boost::make_iterator_property_map(distances.begin(), get(boost::vertex_index, g)))
            .weight_map(weightMap));

        allPredecessors.insert(pair<vertex_t, vector<vertex_t> > ( beginV, predecessors) ) ;
        BGL_FORALL_VERTICES(v, g, Graph)
        {           distancesM (beginV, index_map[v]) = distances[v];         }
    }
    return allPredecessors;
                }


// ====================== Display

void displayGraph( Graph g)
{
    //VertexIndexMap index_map = get(vertex_index, g);
    std::pair<vertex_iter, vertex_iter> vp;
    auto nameMap = boost::get(&VertexProperties::name, g);
    for(unsigned int i =0; i <num_vertices(g); i++)
    {
        cout << i << " : " << nameMap[i] << endl;
    }
    auto weightMap = boost::get(&EdgeProperties::weight, g);
    auto idMap = boost::get(&EdgeProperties::id, g);
    typedef graph_traits<Graph>::edge_iterator edge_iter; // TODO : if used more, put in global typedef
    std::pair<edge_iter, edge_iter> ep;
    edge_iter ei, ei_end;
    for (tie(ei, ei_end) = edges(g); ei != ei_end; ++ei)
    {
        std::cout << idMap[*ei] << " : " <<  nameMap[source(*ei, g)] << "-- " << weightMap[*ei] << " --> " << nameMap[target(*ei, g)] << endl;
    }
}




// ====================== Display

void writeGraph( Graph g,  vector<HyperArc> allHyperA, string & filename)
{
    //VertexIndexMap index_map = get(vertex_index, g);
    ofstream fichier ( filename, ios::out);
    if (fichier)
    {
        std::pair<vertex_iter, vertex_iter> vp;
        auto nameMap = boost::get(&VertexProperties::name, g);
        /*for(unsigned int i =0; i <num_vertices(g); i++)
        {
            fichier << i << " : " << nameMap[i] << endl;
        }*/
        for(auto itH : allHyperA)
           {
               fichier << itH.id << ": " ;
               for(auto itS : itH.substrates)
               {
                   fichier << nameMap[itS] << " ";
               }
               fichier << " ---- " << itH.weight << "----> ";
               for(auto itP : itH.products)
               {
                   fichier << nameMap[itP] << " ";
               }
               fichier << endl;
           }
        auto weightMap = boost::get(&EdgeProperties::weight, g);
        auto idMap = boost::get(&EdgeProperties::id, g);
        typedef graph_traits<Graph>::edge_iterator edge_iter; // TODO : if used more, put in global typedef
        std::pair<edge_iter, edge_iter> ep;
        edge_iter ei, ei_end;
        for (tie(ei, ei_end) = edges(g); ei != ei_end; ++ei)
        {
            fichier << idMap[*ei] << " : " <<  nameMap[source(*ei, g)] << "-- " << weightMap[*ei] << " --> " << nameMap[target(*ei, g)] << endl;
        }
        fichier.close();

    }
    else
    {
        cerr << "Impossible d'ouvrir le fichier " << filename << "!" << endl;
    }
}


void displayHyperArcs( vector<HyperArc> allHyperA, Graph & g)
{
    cout << "= in display hyperHyperarcs" << endl;
    auto nameMap = boost::get(&VertexProperties::name, g);
    for(auto itH : allHyperA)
    {
        cout << itH.id << ": " ;
        for(auto itS : itH.substrates)
        {
            cout << nameMap[itS] << " ";
        }
        cout << " ---- " << itH.weight << "----> ";
        for(auto itP : itH.products)
        {
            cout << nameMap[itP] << " ";
        }
        cout << endl;
    }
}

