#ifndef GRAPH_H
#define GRAPH_H

//=========== BGL =================
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/dijkstra_shortest_paths.hpp>
#include <boost/graph/graphviz.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/graph/iteration_macros.hpp>
#include <boost/graph/properties.hpp>
#include <boost/property_map/property_map.hpp>

// For distance matrix , nb :not sure it is faster than hand implementation, however seems more sure
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/io.hpp>


//========= NameSpace
using namespace boost;
using namespace std;
//============ Graph DEFINITION ===========

/// create a typedef for the Graph type
// Choice for the conytainers :vecS or ListS (listS more space taken, but less price for adding elements)
// in Boost : adjacency_list<OutEdgeList, VertexList, Directed, VertexProperties, EdgeProperties, GraphProperties, EdgeList>
// NB : bidirectional more heavy but we have access to in_edges AND out_edges if needed.
// Properties on edges, vertex and graph can be added if neeeded.

// Definition of edges and Vertices informations :)


struct EdgeProperties
{
    double weight;
    unsigned int id;
    EdgeProperties() : weight(1.0) , id(0){}// Default
    EdgeProperties(double w, unsigned i ) : weight(w), id(i){}
};
struct VertexProperties
{
    string name;
    int role ; //  -1 : source ,  0 : na, 1 target
    VertexProperties() :  name("") {}
    VertexProperties(  string n, int rr) : name(n), role(rr) {}
};



struct graphProperties {};
// RMQ the same properties can be defined for graph : here not done

// stupid hyperArcs structure > ids should not be necessary but I used it for operator < , to do : check if something else is more clever
struct HyperArc
{
    int id;
    double weight;
  //  vector <string> substrates;// could be replace by ref to vertex
  //  vector <string> products;
    vector <int> substrates;
    vector <int> products;
    HyperArc (int i, double w, vector <int> sub, vector<int> prod) : id(i), weight(w), substrates(sub), products(prod){}
};



// ===================== TypeDef ===================


typedef adjacency_list< vecS, vecS, bidirectionalS, VertexProperties,  EdgeProperties>  Graph;
// RMQ: with vecS i am supposed to have automaticaly a vertex_index_t property
// ALso it will be complicated (see documentation) to remove a vertex (but not really our problem is it?
typedef graph_traits<Graph>::vertex_descriptor vertex_t;
typedef graph_traits<Graph>::edge_descriptor edge_t;
typedef graph_traits<Graph>::vertex_iterator vertex_iter;

//My owne propertyMAP
// TODO : no idea what do the star mean!
// ALSO DO NOT WORK WITH const_type, also no idea why
typedef  property_map< Graph, string VertexProperties::* >::type nameMap;
//typedef typename property_map<Graph, name_t>::const_type nameMap;


// Property Map
typedef  property_map<Graph, vertex_index_t>::const_type VertexIndexMap;



typedef numeric::ublas::matrix<double> DistanceMatrix;


//==================== Function Definition ===============
bool operator <(const HyperArc&  x, const HyperArc y) ;


std::vector< vertex_t > getShortestPath(nameMap name, map< vertex_t, vector<vertex_t> > allPredecessors, vertex_t startNode, vertex_t endNode, bool verbose );
std::vector< vertex_t > getShortestPath(map<int, string> name, map< vertex_t, vector<vertex_t> > allPredecessors, int startNode, int endNode, bool verbose );

map< vertex_t, vector<vertex_t> > computeDijkstra( Graph g, numeric::ublas::matrix<double> & distancesM  );


void displayGraph( Graph g) ;
void displayHyperArcs( vector<HyperArc> allHyperA, Graph & g);

void writeGraph( Graph g,  vector<HyperArc> allHyperA, string & filename);

#endif
