
#include <iostream>
#include <set>
#include <queue>
#include <iomanip>
#include <boost/graph/graph_concepts.hpp>
#include <boost/math/special_functions/factorials.hpp>
#include <limits>
#include <map>


#include <thread>
#include <mutex>

#include "Graph.h"


#define INF std::numeric_limits < double >::max()
//Shortest Path: using matrix or arrays?
#define SP(a , b)  sp(a,b)
//#define SP(a , b)  sp[a][b]; 
//DistanceMatrix

using namespace std;

typedef double Weight;
typedef std::set < int >Vertex_Set;
typedef pair < Vertex_Set, int > Key;
typedef pair < int, int > PathPair;





//====================== Global Variables ; already declared in main...================
// Will be fill using progam options
extern int g_max_tentacules;
extern int g_verbose ;
// 0 = no output (except errors)
// 1 = progress (every x step, see below), and better solutions
// 2 = summary of each step
// 3 = detail of each step, including steiner trees
// 4 = everything
extern int g_nb_threads;
extern int g_step_print_progress;
extern bool g_check_steiner_tree;
extern bool g_dry_run;//DRY_RUN: do not compute the solution, just enumerate the feasible subsets of hyperedges.
extern bool g_use_lowerbounds ;//USE_LOWERBOUNDS: compute lower bounds for each H, and skip those with l.b. higher than current optimal.
extern bool g_keep_stats ; //KEEP_STATS: do statistics over the run, takes additional synchronization.
extern bool g_reset_from_earlier; // skip the first sets in the search space, to reset computation where it was left by an earlier run.
extern vector<int> g_reset_from;
extern string outDir;





// This is a struct to be able to write the solution.
// getFullSolution
struct writeSolution {
    map< vertex_t, vector<vertex_t> >  allPredecessors;
    nameMap name;
    vector <HyperArc> allHyperA;
    //writeSolution(){};
    writeSolution ( map< vertex_t, vector<vertex_t> >  aP, nameMap n, vector <HyperArc> aH ) :  allPredecessors(aP), name(n), allHyperA(aH) {};
};

struct SteinerValue {
   Weight value;
   bool singlePath;
   int steinerVertex;
   Vertex_Set leftPartition;
   Vertex_Set rightPartition;
};


struct HyperSteinerValue {
   Weight value;
   bool singleTree;
   int source;   
   Vertex_Set leftPartition;
   Vertex_Set rightPartition;
   int rank;
};

struct HyperSteinerElement {
   int source;
   set<int> cover;
   int rank;
   HyperSteinerElement(int s, set<int> c, int r) : source(s), cover(c), rank(r) {}
};

struct HyperSteinerForest {
   Weight value;
   vector<HyperSteinerElement> steinerTrees;
   set<PathPair> solutionPaths;
   vector<int> hyperId;

};

typedef std::map < Key, SteinerValue > Steiner;
typedef pair < Key, SteinerValue > SteinerPair;


typedef std::map < Vertex_Set, HyperSteinerValue > HyperSteiner;
typedef pair < Vertex_Set, HyperSteinerValue > HyperSteinerPair;


int test();
Steiner steiner(int *T, int TLength, int nVertex, DistanceMatrix& sp);
HyperSteinerForest hyperSteiner( vector<int>& S, vector<int>& T, int nVertex, vector<HyperArc>& hyperArcs, DistanceMatrix& sp);
HyperSteinerForest parallelHyperSteiner( vector<int>& S, vector<int>& T, int nVertex, vector<HyperArc>& hyperArcs, DistanceMatrix& sp,  const writeSolution & infoSol);

void printSteiner(Steiner X, const vector<int> T, int nVertex);
void printHyperSteinerForest(HyperSteinerForest F);

void getFullSolution(HyperSteinerForest F, string fileName, const writeSolution & infoSol);
