/*
 * main.cpp
 *
 *  Created on: 30 juin 2014
 *      Author: ajulien
 */
#include <iostream>                  // for cout
#include <utility>                   // for pair
#include <algorithm>                 // for for_each
//#include <string>
#include <algorithm>                // for next permutation
#include <ctime>                    //mesuring time

#include <boost/program_options.hpp>

#include <boost/tokenizer.hpp>
#include <boost/lexical_cast.hpp>
//================ project files =========
#include "ReadWrite.h"

#include "Steiner.h"


//========= NameSpace
using namespace boost;
using namespace std;

// To be able to use program options: => from the boost documentations
namespace po = boost::program_options;



//====================== Global Variables ================
// Will be fill using progam options
int g_max_tentacules;
int g_verbose ;
// 0 = no output (except errors)
// 1 = progress (every x step, see below), and better solutions
// 2 = summary of each step
// 3 = detail of each step, including steiner trees
// 4 = everything
int g_nb_threads;
int g_step_print_progress;
bool g_check_steiner_tree;
bool g_dry_run;//DRY_RUN: do not compute the solution, just enumerate the feasible subsets of hyperedges.
bool g_use_lowerbounds ;//USE_LOWERBOUNDS: compute lower bounds for each H, and skip those with l.b. higher than current optimal.
bool g_keep_stats ; //KEEP_STATS: do statistics over the run, takes additional synchronization.
bool g_reset_from_earlier; // skip the first sets in the search space, to reset computation where it was left by an earlier run.
vector<int> g_reset_from;
string outDir;


/*!
 * Usage :
 *
 * exectuable fileWithArc fileHyperArc, maxNbHyperArcs, nbThreads, resetFromEarlier, Resert_From, verboseLevel, dryRun,
 */
int main(int argc, char** argv )
{
    //=================================================================================================
    //              Reading program options
    //==================================================================================================
    string fileName, hyperNames;
    try {
        // Declare the supported options.
        po::options_description desc("Allowed options");

        //Default Values?
        //   int k_in_solution, nb_threads,  verbose, step_print_progress;
        //   bool reset_from_earlier, dry_run, use_lower_bounds,keep_stats, check_steiner_tree;
        string reset_fromS;
        // Describing the allowed parameters and fixing default values
        desc.add_options()
                                ("help", "produce help message")
                                ("arcFile", po::value<string>(), "arc file")
                                ("hyperArcFile", po::value<string>(), "hyperArcFile")
                                ("k", po::value<int>(&g_max_tentacules)->default_value(3), "max number of hyperarcs")
                                ("nbThreads", po::value<int>(&g_nb_threads)->default_value(2), "number of threads to use")
                                ("reset", po::value<bool>(&g_reset_from_earlier)->default_value(false), "Boolean, whether to start from the beginning or from the starting point")
                                ("reset_From", po::value<string>(&reset_fromS)->default_value(""), "comma separated string containing the positions from which each threads should start, e.g: 1,10,15,20")
                                ("verboseLevel", po::value<int>(&g_verbose)->default_value(1), "verbose level ")
                                ("dryrun", po::value<bool>(&g_dry_run)->default_value(false), "dry Run")
                                ("USE_LB", po::value<bool>(&g_use_lowerbounds)->default_value(true), "useLowerBounds")// true
                                ("KeepStats", po::value<bool>(&g_keep_stats)->default_value(true), "do statistics over the run, takes additional synchronization.")
                                ("stepP", po::value<int>(&g_step_print_progress)->default_value(1000), "STEP_PRINT_PROGRESS")// 1000
                                ("CHECK_ST", po::value<bool>(&g_check_steiner_tree)->default_value(false), "check steiner tree")// false
                                ("outDir", po::value<string>(&outDir)->default_value("resultsMultipus"))
                                ;

        po::variables_map vm;
        po::store(po::parse_command_line(argc, argv, desc), vm);
        po::notify(vm);


        if (vm.count("help")) {
            cout << desc << "\n";
            return 0;
        }

        // Checking the arguments
        if (! (vm.count("arcFile") && vm.count("hyperArcFile") )) {
            cout << "Required arcFile and hyperArcFile options ";
            return 1;
        }
        else {
            cout << "Reading arcs, S and T from : " << vm["arcFile"].as<string>() << " and hyperarcs from: " << vm["hyperArcFile"].as<string>() << endl;
            fileName = vm["arcFile"].as<string>();
            hyperNames = vm["hyperArcFile"].as<string>() ;

        }
        //Checking wheter the new starting points are in accordance with the number of threads!
        if (vm["reset"].as<bool>() ) {
            stringstream stream(reset_fromS);
            char_separator<char> sep(", ");

            tokenizer<char_separator<char>> tokens(reset_fromS, sep);


            for(auto & it : tokens)
            {
                // try {
                //     int x = lexical_cast<int>( it );
                // } catch( boost::bad_lexical_cast const& ) {
                //     std::cout << "Error: input string was not valid" << std::endl;
                //}
                // reset_from.push_back(x);
                g_reset_from.push_back( boost::lexical_cast<int>(it) );
            }
            // We need to check that the number of threads given is ok
            if( (int) g_reset_from.size() != g_nb_threads)
            {
                cout << " reset_from should be the size than the number of threads!!" << endl;
                return 1;
            }
        }
    }
    /*  catch(exception& excep) {
        cerr << "error: " << excep.what() << "\n";
        return 1;
    }*/
    catch(...) {
        cerr << "Exception of unknown type!\n";
    }
    // Summary of the options
    cout << " dsh_main launched with k "<< g_max_tentacules << "; verbose  " << g_verbose << "; " << g_nb_threads <<" threads;  print every " << g_step_print_progress << "; dry_run" << g_dry_run;
    cout <<"; lowerbounds " << g_use_lowerbounds<< "; keeping stats  " << g_keep_stats<< endl;
    int countingThreads = 0;
    if(g_reset_from_earlier)
    {
        cout << " resetting from previous solution \n [ ";
        for(auto nbTh : g_reset_from)
        {
            cout << countingThreads << " : " << nbTh << " , " ;
            ++countingThreads;
        }
        cout << "] "<< endl;
    }

    //=================================================================================================
    //              Read graph files
    //==================================================================================================

    //=== Compound Graph
    Graph myCompoundG =  readAsCompoundGraph(fileName, hyperNames);
    vector<HyperArc> mm;
    //string toto2 = "compoundG.txt";
    //  writeGraph(myCompoundG, mm,toto2);
    numeric::ublas::matrix<double> distanceGCompound ( num_vertices(myCompoundG), num_vertices(myCompoundG) );
    map< vertex_t, vector<vertex_t> > allPredecessorsCompoundG =   computeDijkstra(myCompoundG, distanceGCompound);

    // nameMap nameG = boost::get(&VertexProperties::name, myCompoundG);
    /*  ofstream fichierG("matrixCompoundG.txt", ios::out);  //writing solution
    // Show distances?
    if (fichierG)
    {
        for(unsigned int i = 0  ; i < distanceGCompound.size1(); i++)
        {
            for(unsigned int j = 0  ; j < distanceGCompound.size2(); j++)
            {
                fichierG  << "("<< nameG[i] << "," << nameG[j] << ") : " ;
                if (  distanceGCompound(i, j)<INF)
                    fichierG  << distanceGCompound(i, j) << "\n" ;
                else
                    fichierG  << " INF \n" ;
            }
        }
        fichierG.close();
    }
     */

    //===  Write the graph
    //writeGraphViz("compoundG.dot", myCompoundG);


    //==HyperGraph : arcs + hyperarcs

    map<string, vertex_t> checkNodes; // TODO: should not be needed. Easier for the moment!
    //========= READING =============
    map<string, int> nodeIdMap;
    Graph g = readGraphFile(fileName, checkNodes);


    vector<int> sources ; // vertex that are sources
    vector<int> targets ; // vertex that are targets
    auto roleMap = boost::get(&VertexProperties::role, g);
    auto nameM = boost::get(&VertexProperties::name, g);
    for ( unsigned int i=0; i<num_vertices(g); i++) {
        nodeIdMap.insert(pair<string,int>(nameM[i], i));
        if (roleMap[i] == -1) { sources.push_back(i);}
        else if (roleMap[i] == 1) { targets.push_back(i);}

    }    
    vector <HyperArc> allHyperA = readHyperArcsFile(hyperNames, nodeIdMap, g);
    string toto = "myGraph.txt";
    writeGraph(g, allHyperA, toto );
    //  displayHyperArcs(allHyperA, g);

    //=================================================================================================
    //              Computing Distance Matrix
    //==================================================================================================
    numeric::ublas::matrix<double> matrixDistance ( num_vertices(g), num_vertices(g) );
    map< vertex_t, vector<vertex_t> > allPredecessors =   computeDijkstra(g, matrixDistance);

    nameMap name = boost::get(&VertexProperties::name, g);
    /*  ofstream fichier("matrix.txt", ios::out);  //writing solution
    // Show distances?
    if (fichier)
    {
        for(unsigned int i = 0  ; i < matrixDistance.size1(); i++)
        {
            for(unsigned int j = 0  ; j < matrixDistance.size2(); j++)
            {
                fichier  << "("<< name[i] << "," << name[j] << ") : " ;
                if (  matrixDistance(i, j)<INF)
                    fichier  << matrixDistance(i, j) << "\n" ;
                else
                    fichier  << " INF \n" ;
            }
        }
        fichier.close();
    }

     */
    // Testing shortest path

    VertexIndexMap index_map = get(vertex_index, g);


    //== Write the graph
    // writeGraphViz("graph", g);

    //== Checking
    // VertexIndexMap index_map = get(vertex_index, g);
    cout << "  \n sources: " ;
    for(auto i : sources){ cout <<  nameM[ index_map[i]] << " \t ";}
    cout << " \n targets: " ;
    for(auto i : targets){ cout <<  nameM[ index_map[i]] << " \t ";}
    cout << endl;



    //=================================================================================================
    //              Computing the solutions
    //==================================================================================================

    const writeSolution myInfoSol (allPredecessors, name, allHyperA);
    HyperSteinerForest  F = parallelHyperSteiner( sources, targets, num_vertices(g), allHyperA, matrixDistance, myInfoSol);
    //=================================================================================================
    //              Writing the best solution
    //==================================================================================================


    printHyperSteinerForest(F) ;
    getFullSolution(  F, outDir +  "/BestSolution", myInfoSol );


    return 0;
}




